#include "robot.h"

Robot::Robot(void)
: m_ConnectThread( NULL )
, m_Port( 0 )
, m_wallHit( false )
, m_robotInSight( false )
, m_foodFound( false )
, m_attacked( false )
, m_actStupid( false )
{
}

Robot::~Robot(void)
{
    // Stop and delete the connect thread
    if( m_ConnectThread )
    {
        m_ConnectThread->Stop();
        m_ConnectThread->WaitUntilTerminate();

        delete m_ConnectThread;
        m_ConnectThread = NULL;
    }
}

bool Robot::initialize(unsigned short port)
{
    // Create the ConnectThread
    if( m_ConnectThread )
    {
        m_ConnectThread->Stop();
        m_ConnectThread->WaitUntilTerminate();

        delete m_ConnectThread;
        m_ConnectThread = NULL;
    }

    m_ConnectThread = new ConnectThread( this, port );

    if( NULL == m_ConnectThread )
    {
        return( false );
    }

    if( m_ConnectThread )
    {
        m_ConnectThread->Start();
    }

    return( true );
}

////////////////////////////////////////////////////////////////////
// Robot::ConnectThread member functions

Robot::ConnectThread::ConnectThread(Robot * r, unsigned short port)
: CThread()
, m_bStop( false )
, m_Port( port )
, m_Robot( r )
, m_ClientSocket( SOCKET_ERROR )
, m_hasTurn( false )
, m_hMutex( NULL )
{
}

void Robot::ConnectThread::OnInitInstance()
{
    int Error;

    if( NULL == m_hMutex )
    {
        m_hMutex = ::CreateMutex( NULL, FALSE, NULL );
    }

    if( NULL == m_hMutex )
    {
        Stop();
        return;
    }

    if( INVALID_SOCKET == m_ClientSocket )
    {
        m_ClientSocket = ::socket( AF_INET, SOCK_STREAM, 0 );
    }

    if( INVALID_SOCKET != m_ClientSocket )
    {
        SOCKADDR_IN saTemp;
        saTemp.sin_family = AF_INET;
        saTemp.sin_port = ::htons( m_Port );

        unsigned long ulAddress = ::inet_addr("127.0.0.1");
        saTemp.sin_addr.S_un.S_addr = ulAddress;
/*
        unsigned long ulAddress = ::inet_addr("127.0.0.1");

        if( INADDR_NONE == ulAddress )
        {
            pHost = gethostbyname (ServerName);
            if (NULL == pHost)
            {
                HandleError("Unable to resolve host name ");
            }
            memcpy((char FAR *)&ulAddress, pHost->h_addr, pHost->h_length);
        }
*/
        if( SOCKET_ERROR != ::connect( m_ClientSocket, (PSOCKADDR)&saTemp, sizeof(saTemp) ) )
        {
            registerCommand rc(20, 20);

            if( SOCKET_ERROR == ::send( m_ClientSocket, (char *)&rc, sizeof(rc), 0 ) )
            {
                Error = ::WSAGetLastError();
                Stop();
            }
        }
        else
        {
            Error = ::WSAGetLastError();
            Stop();
        }
    }
    else
    {
        Error = ::WSAGetLastError();
        Stop();
    }
}

void Robot::ConnectThread::OnRunning(void)
{
    int Error;                  // Only needed for debug purposes
    BasicCommand cmd;           // Every command starts with a header
    int BytesRead;

    while( !m_bStop )
    {
        // Every time a poll command is received, the robot is allowed
        //   to do one action, i.e. send a command to the server.
        BytesRead = ::recv( m_ClientSocket, (char *)&cmd, sizeof(cmd), MSG_PEEK );

        // Check if the socket communication succeeded
        if( (SOCKET_ERROR == BytesRead) || (0 == BytesRead) )
        {
            Error = ::WSAGetLastError();
            Stop();
            continue;
        }

        // Check the version
        if( theVersion != cmd.m_version )
        {
            Stop();
            continue;
        }

        doTurn( cmd.m_command );
    }
}

DWORD Robot::ConnectThread::OnExitInstance()
{
    if( m_hMutex )
    {
        ::CloseHandle( m_hMutex );

        m_hMutex = NULL;
    }

    QueuedCommandPair p;

    while( !m_QueuedCommands.empty() )
    {
        p = m_QueuedCommands.front();
        m_QueuedCommands.pop();

        if( p.first )
        {
            delete p.first;
        }
    }

    return( 0 );
}

void Robot::ConnectThread::Stop(void)
{
    int Error;

    if( !m_bStop )
    {
        m_bStop = true;

        if( INVALID_SOCKET != m_ClientSocket )
        {
            if( SOCKET_ERROR == ::shutdown( m_ClientSocket, 2 ) )
            {
                Error = ::WSAGetLastError();
            }

            if( SOCKET_ERROR == ::closesocket( m_ClientSocket ) )
            {
                Error = ::WSAGetLastError();
            }
        }
    }
}

void Robot::ConnectThread::doTurn(CommandType & cmdType)
{
    int Error;
    int BytesRead = 0;

    // All possible commands and events that can arrive (can be optimized)
    pollCommand pollCmd;
    infoCommand infoCmd;
    onHitWallCommand onHitWallCmd;
    onRobotInSightCommand onRobotInSightCmd;
    onDeadCommand onDeadCmd;
    onFoodFoundCommand onFoodFoundCmd;
    onAttackedCommand onAttackedCmd;
    generalInfoCommand generalInfoCmd;

    // Receive the complete command or event
    switch( cmdType )
    {
    case cmdPoll:
        BytesRead = ::recv( m_ClientSocket, (char *)&pollCmd, sizeof(pollCmd), 0 );
        break;
    case cmdInfo:
        BytesRead = ::recv( m_ClientSocket, (char *)&infoCmd, sizeof(infoCmd), 0 );
        break;
    case cmdHitWall:
        BytesRead = ::recv( m_ClientSocket, (char *)&onHitWallCmd, sizeof(onHitWallCmd), 0 );
        break;
    case cmdRobotInSight:
        BytesRead = ::recv( m_ClientSocket, (char *)&onRobotInSightCmd, sizeof(onRobotInSightCmd), 0 );
        break;
    case cmdDead:
        BytesRead = ::recv( m_ClientSocket, (char *)&onDeadCmd, sizeof(onDeadCmd), 0 );
        break;
    case cmdFoodFound:
        BytesRead = ::recv( m_ClientSocket, (char *)&onFoodFoundCmd, sizeof(onFoodFoundCmd), 0 );
        break;
    case cmdAttacked:
        BytesRead = ::recv( m_ClientSocket, (char *)&onAttackedCmd, sizeof(onAttackedCmd), 0 );
        break;
    case cmdGeneralInfo:
        BytesRead = ::recv( m_ClientSocket, (char *)&generalInfoCmd, sizeof(generalInfoCmd), 0 );
        break;
    }

    // Check if the socket communication succeeded
    if( (SOCKET_ERROR == BytesRead) || (0 == BytesRead) )
    {
        Error = ::WSAGetLastError();
        Stop();
        return;
    }

    switch( cmdType )
    {
    case cmdPoll:
        m_Robot->onTurn();                  // Call "callback" function
        m_hasTurn = true;
        break;
    case cmdInfo:                           // Retrieve the info
        m_Robot->m_Properties.m_ID = infoCmd.m_id;
        m_Robot->m_Properties.m_food = infoCmd.m_status.m_food;
        m_Robot->m_Properties.m_life = infoCmd.m_status.m_life;
        m_Robot->m_Properties.m_heading = infoCmd.m_status.m_heading;
        break;
    case cmdHitWall:
        m_Robot->onHitWall();               // Call "callback" function
        break;
    case cmdRobotInSight:
        m_Robot->onRobotInSight();          // Call "callback" function
        break;
    case cmdDead:
        m_Robot->onDead();                  // Call "callback" function
        break;
    case cmdFoodFound:
        m_Robot->onFoundFood();             // Call "callback" function
        break;
    case cmdAttacked:
        m_Robot->onAttacked();              // Call "callback" function
        break;
    case cmdGeneralInfo:
        m_Robot->m_Nr = generalInfoCmd.m_Nr;
        break;
    }

    if( m_hasTurn )
    {
        SendCommand();                      // Send the next command in the queue

        m_hasTurn = false;
    }
}

bool Robot::ConnectThread::QueueCommand(char * cmd, int size)
{
    QueuedCommandPair p;

    p.first = new char[size];

    if( NULL == p.first )
    {
        return( false );
    }

    ::memcpy( p.first, cmd, size );
    p.second = size;

    ::WaitForSingleObject( m_hMutex, INFINITE );

    m_QueuedCommands.push( p );

    ::ReleaseMutex( m_hMutex );

    return( true );
}

bool Robot::ConnectThread::SendCommand(void)
{
    if( !m_hasTurn )
    {
        return( false );
    }

    int Error;
    int BytesSent = SOCKET_ERROR;

    QueuedCommandPair p;
    skipTurnCommand skip;           // Needed in case the queue is empty

    ::WaitForSingleObject( m_hMutex, INFINITE );

    if( m_QueuedCommands.empty() )
    {
        ::ReleaseMutex( m_hMutex );

        p.first = (char *)&skip;
        p.second = sizeof(skip);

        BytesSent = ::send( m_ClientSocket, p.first, p.second, 0 );
    }
    else
    {
        p = m_QueuedCommands.front();
        m_QueuedCommands.pop();

        ::ReleaseMutex( m_hMutex );

        if( p.first )
        {
            BytesSent = ::send( m_ClientSocket, p.first, p.second, 0 );

            delete p.first;
        }
        else
        {
            return( false );            // Should never occur
        }
    }

    if( SOCKET_ERROR == BytesSent )
    {
        Error = ::WSAGetLastError();
        Stop();

        return( false );
    }

    return( true );
}

////////////////////////////////////////////////////////////////////
// Robot command functions

void Robot::rest(void)
{
    bool sent;

    if( m_ConnectThread )
    {
        restCommand r;

        sent = m_ConnectThread->QueueCommand( (char *)&r, sizeof(r) );
    }
}

void Robot::moveForward(void)
{
    bool sent;

    if( m_ConnectThread )
    {
        moveForwardCommand mf;

        sent = m_ConnectThread->QueueCommand( (char *)&mf, sizeof(mf) );
    }
}

void Robot::moveBackward(void)
{
    bool sent;

    if( m_ConnectThread )
    {
        moveBackwardCommand mb;

        sent = m_ConnectThread->QueueCommand( (char *)&mb, sizeof(mb) );
    }
}

void Robot::turnLeft(void)
{
    bool sent;

    if( m_ConnectThread )
    {
        turnLeftCommand tl;

        sent = m_ConnectThread->QueueCommand( (char *)&tl, sizeof(tl) );
    }
}

void Robot::turnRight(void)
{
    bool sent;

    if( m_ConnectThread )
    {
        turnRightCommand tr;

        sent = m_ConnectThread->QueueCommand( (char *)&tr, sizeof(tr) );
    }
}

void Robot::eat(void)
{
    bool sent;

    if( m_ConnectThread )
    {
        eatCommand eat;

        sent = m_ConnectThread->QueueCommand( (char *)&eat, sizeof(eat) );
    }
}

void Robot::fight(void)
{
    bool sent;

    if( m_ConnectThread )
    {
        fightCommand fight;

        sent = m_ConnectThread->QueueCommand( (char *)&fight, sizeof(fight) );
    }
}

////////////////////////////////////////////////////////////////////
// Robot "callback" functions

void Robot::onTurn(void)
{
    if( m_actStupid )
    {
        static bool forward = false;    // Keep track of the last move, was it forward?

        if( m_wallHit )
        {
            if( forward )
            {
                moveBackward();
                forward = false;
            }
            else
            {
                moveForward();
                forward = true;
            }

            m_wallHit = false;
        }
        else
        {
            if( m_attacked )
            {
                moveBackward();
                forward = false;
            }
            else
            {
                if( m_robotInSight )
                {
                    if( (m_Properties.m_life > 50) && (m_Properties.m_food > 50) )
                    {
                        fight();
                    }
                    else
                    {
                        moveBackward();
                        forward = false;
                    }

                    m_robotInSight = false;
                }
                else
                {
                    if( m_foodFound && (m_Properties.m_food < valMaxFood) )
                    {
                        eat();
                    }
                    else
                    {
                        if( m_Properties.m_life < 50 )
                        {
                            rest();
                        }
                        else
                        {
                            int count = ::rand() % 7;

                            switch( count )
                            {
                            case 0: turnRight();   break;
                            case 1: rest();        break;
                            case 2: moveForward(); forward = true; break;
                            case 3: rest();        break;
                            case 4: turnLeft();    break;
                            case 5: rest();        break;
                            case 6: moveForward(); forward = true; break;
                            }
                        }
                    }

                    m_foodFound = false;
                }
            }
        }

        m_attacked = false;
    }

    return;
}

void Robot::onHitWall(void)
{
    m_wallHit = true;
    return;
}

void Robot::onRobotInSight(void)
{
    m_robotInSight = true;
    return;
}

void Robot::onDead(void)
{
    return;
}

void Robot::onFoundFood(void)
{
    m_foodFound = true;
    return;
}

void Robot::onAttacked(void)
{
    m_attacked = true;
    return;
}
